/* @(#) tsubAA.c 2.2 2015/10/27 -- speed propagation */

/* tsubAA.c - Transformation Subroutines For Aperture (pinhole collimator) angle      */
/* These equations are for pitch (vertical) and yaw (horizontal) angular adjustment   */
/* of apertures with two push blocks. The aperture starts to rotate once the push     */
/* motor reaches the prism. The Pitch rotation is independent while the Yaw  rotation */
/* depends on the Pitch because the more Pitch is pushed, the later the Yae pusher    */
/* reaches the Yaw prism. The equations are approximate for small rotation angles.    */

#include	<stdlib.h>
#include	<stdio.h>
#include	<string.h>
#include	<math.h>

#include	<dbDefs.h>
#include	<tsubRecord.h>
#include	<dbCommon.h>
#include	<recSup.h>
#include	<epicsExport.h>
#include	<registryFunction.h>

volatile int tsubAADebug = 0;
#define TSUB_MESSAGE	logMsg
#define TSUB_TRACE(level,code) { if ( (pRec->tpro == (level)) || (tsubAADebug == (level)) ) { code } }

/* ===========================================
 * tsubAPang - Aperture Initialization
 */
static long tsubAPang (struct tsubRecord *pRec) {
 	return (0);
}


/* ===========================================
 * tsubAPangSync
 *	oa = m1:RqsPos
 *	ob = m2:RqsPos
 *	a  = m1:ActPos
 *	b  = m2:ActPos
 */
static long tsubAPangSync (struct tsubRecord *pRec) {
	pRec->oa = pRec->a;
	pRec->ob = pRec->b;
	return (0);
}


/* ===========================================
 * tsubAPangMtr - Aperture Angle: Motors To Drives and Axes
 *	oa = x1		(Pitch angle)
 *	ob = x2		(Yaw angle)
 *	oa1 = d1	(Pitch pusher drive)
 *	ob1 = d2	(Yaw pusher drive)
 *      oc = Sp         (Pitch pusher engaged length)
 *      od = Sy         (Yaw pusher engaged length at Sp=0)
 *      oe = Sw         (Yaw pusher engaged length real, with account for Sp)
 *	a = m1		(Pitch pusher motor)
 *	b = m2		(Yaw pusher motor)
 *      i = Lp		(Pitch frame level arm: distance between Pitch axis and Pitch prism)
 *      j = Ly		(Yaw frame level vertical arm: distance between Pitch axis and Yaw prism)
 *      k = Lw		(Yaw frame level horizontal arm: distance between Yaw axis and Yaw prism)
 *      l = Tp		(Pitch drive position when it reaches its prism and becomes ready to push)
 *      m = Ty		(Yaw drive position when it reaches its prism and becomes ready to push -- when the Yaw-prism is not displaced by Pitch pusher)
 *	a0 = d1:Offset
 *	a1 = d1:Scale
 *      a2 = d1:ActPos  (Pitch pusher drive actual position)
 *	a3 = x1:Offset	(initial Pitch angle when nothing is pushed)
 *	b0 = d2:Offset
 *	b1 = d2:Scale
 *      b2 = d2:ActPos	(Yaw pusher drive actual position)
 *	b3 = x2:Offset	(initial Yaw angle when nothing is pushed)
 *	nla =		(1=wo/Offsets)[rel,vel] (0=w/Offsets)[abs,pos]
 */
static long tsubAPangMtr (struct tsubRecord *pRec) {
	double Lp, Ly, Lw, Tp, Ty, Sp, Sy, Sw, Sp0, Sy0, Sw0,
               d1_offset, d1_scale, d1_act, x1_offset,
               d2_offset, d2_scale, d2_act, x2_offset;
	Lp        = pRec->i;
	Ly        = pRec->j;
	Lw        = pRec->k;
	Tp        = pRec->l;
	Ty        = pRec->m;
	d1_offset = pRec->a0;
	d1_scale  = pRec->a1;
	d1_act    = pRec->a2;
	x1_offset = pRec->a3;
	d2_offset = pRec->b0;
	d2_scale  = pRec->b1;
	d2_act    = pRec->b2;
	x2_offset = pRec->b3;

	if (Lp == 0.0 || Ly  == 0.0 || Lw == 0.0) {
		printf ("tsubAPangMtr: exit on zero Lp, Ly, or Lw\n");
		return (-1);
	}

	if (pRec->nla == 0.0)					   			/* absolute motion */
	{
		pRec->oa1 = pRec->a * d1_scale + d1_offset;         			/* d1 = m1*scale1+offset1 */
		pRec->ob1 = pRec->b * d2_scale + d2_offset;         			/* d2 = m2*scale2+offset2 */
		Sp        = pRec->oa1 - Tp;						/* Sp = d1 - pitch_engage */
		Sy        = pRec->ob1 - Ty;						/* Sy = d2 - yaw_engage */
		if (Sp > 0.0) {Sw = Sy - Sp * Ly / Lp;}
		else          {Sw = Sy;}
		pRec->oc  = Sp;
		pRec->od  = Sy;
		pRec->oe  = Sw;
                if (Sp > 0.0) {pRec->oa = x1_offset + 1000. * Sp / Lp;}			/* x1 = x1_offset + 1000*Sp/Lp */
		else          {pRec->oa = x1_offset;}					/* x1 = x1_offset */
		if (Sw > 0.0) {pRec->ob = x2_offset + 1000. * Sw / Lw;}			/* x2 = x2_offset + 1000*Sw/Lp */
		else          {pRec->ob = x2_offset;}					/* x2 = x2_offset */
	}
	else                                                       			/* relative motion */
	{
		pRec->oa1 = pRec->a * d1_scale;                    			/* dd1 = dm1*scale1 */
		pRec->ob1 = pRec->b * d2_scale;                    			/* dd2 = dm2*scale2 */
		Sp0 = d1_act - Tp;		  if (Sp0 < 0.0) Sp0 = 0.0;		/* Sp0 = d1 - pitch_engage */
		Sy0 = d2_act - Ty;		  if (Sy0 < 0.0) Sy0 = 0.0;		/* Sy0 = d2 - yaw_engage */
		Sw0 = Sy0 - Sp0 * Ly / Lp;	  if (Sw0 < 0.0) Sw0 = 0.0;
		Sp  = (d1_act + pRec->oa1) - Tp;  if (Sp  < 0.0) Sp  = 0.0;		/* Sp = (d1+dd1) - pitch_engage */
		Sy  = (d2_act + pRec->ob1) - Ty;  if (Sy  < 0.0) Sy  = 0.0;		/* Sy = (d2+dd2) - yaw_engage */
		Sw  = Sy - Sp * Ly / Lp;          if (Sw  < 0.0) Sw  = 0.0;
		pRec->oa = 1000. * (Sp - Sp0) / Lp;					/* dx1 = 1000*dSp/Lp */
		pRec->ob = 1000. * (Sw - Sw0) / Lw;					/* dx2 = 1000*dSw/Lw */
	}
	return (0);
}

/* ===========================================
 * tsubAPangDrv - Aperture Angle: Drives to Motors and Axes
 *	oa = x1		(Pitch angle)
 *	ob = x2		(Yaw angle)
 *	oa0 = m1	(Pitch pusher motor)
 *	ob0 = m2	(Yaw pusher motor)
 *	a = d1		(Pitch pusher drive)
 *	b = d2		(Yaw pusher drive)
 *      i = Lp		(Pitch frame level arm: distance between Pitch axis and Pitch prism)
 *      j = Ly		(Yaw frame level vertical arm: distance between Pitch axis and Yaw prism)
 *      k = Lw		(Yaw frame level horizontal arm: distance between Yaw axis and Yaw prism)
 *      l = Tp		(Pitch drive position when it reaches its prism and becomes ready to push)
 *      m = Ty		(Yaw drive position when it reaches its prism and becomes ready to push -- when the Yaw-prism is not displaced by Pitch pusher)
 *	a0 = d1:Offset
 *	a1 = d1:Scale
 *      a2 = d1:ActPos  (Pitch pusher drive actual position)
 *	a3 = x1:Offset	(initial Pitch angle when nothing is pushed)
 *	b0 = d2:Offset
 *	b1 = d2:Scale
 *      b2 = d2:ActPos	(Yaw pusher drive actual position)
 *	b3 = x2:Offset	(initial Yaw angle when nothing is pushed)
 *	nla = 		(1=wo/Offsets)[rel,vel] (0=w/Offsets)[abs,pos]
 */
static long tsubAPangDrv (struct tsubRecord *pRec) {
	double Lp, Ly, Lw, Tp, Ty, Sp, Sy, Sw, Sp0, Sy0, Sw0,
               d1_offset, d1_scale, d1_act, x1_offset,
               d2_offset, d2_scale, d2_act, x2_offset;
	Lp        = pRec->i;
	Ly        = pRec->j;
	Lw        = pRec->k;
	Tp        = pRec->l;
	Ty        = pRec->m;
	d1_offset = pRec->a0;
	d1_scale  = pRec->a1;
	d1_act    = pRec->a2;
	x1_offset = pRec->a3;
	d2_offset = pRec->b0;
	d2_scale  = pRec->b1;
	d2_act    = pRec->b2;
	x2_offset = pRec->b3;

	if (d1_scale == 0.0 || d2_scale == 0.0) {
		printf ("tsubAPangDrv: exit on zero mtr to drv scale\n");
		return (-1);
	}
	if (Lp == 0.0 || Ly  == 0.0 || Lw == 0.0) {
		printf ("tsubAPangDrv: exit on zero Lp, Ly, or Lw\n");
		return (-1);
	}

	if (pRec->nla == 0.0)					   			/* absolute motion */
	{
		pRec->oa0 = (pRec->a - d1_offset) / d1_scale;       			/* m1=(d1 -offset1)/scale1 */
		pRec->ob0 = (pRec->b - d2_offset) / d2_scale;       			/* m2=(d2'-offset2)/scale2 */
		Sp        = pRec->a - Tp;						/* Sp = d1 - pitch_engage */
		Sy        = pRec->b - Ty;						/* Sy = d2 - yaw_engage */
		if (Sp > 0.0) {Sw = Sy - Sp * Ly / Lp;}
		else          {Sw = Sy;}
                if (Sp > 0.0) {pRec->oa = x1_offset + 1000. * Sp / Lp;}			/* x1 = x1_offset + 1000*Sp/Lp */
		else          {pRec->oa = x1_offset;}					/* x1 = x1_offset */
		if (Sw > 0.0) {pRec->ob = x2_offset + 1000. * Sw / Lw;}			/* x2 = x2_offset + 1000*Sw/Lp */
		else          {pRec->ob = x2_offset;}					/* x2 = x2_offset */
	}
	else                                                       			/* relative motion */
	{
		pRec->oa0 = pRec->a / d1_scale;                  			/* dm1=dd1/scale1 */
		pRec->ob0 = pRec->b / d2_scale;                  			/* dm2=dd2/scale2 */
		Sp0 = d1_act - Tp;		  if (Sp0 < 0.0) Sp0 = 0.0;		/* Sp0 = d1 - pitch_engage */
		Sy0 = d2_act - Ty;		  if (Sy0 < 0.0) Sy0 = 0.0;		/* Sy0 = d2 - yaw_engage */
		Sw0 = Sy0 - Sp0 * Ly / Lp;	  if (Sw0 < 0.0) Sw0 = 0.0;
		Sp  = (d1_act + pRec->a) - Tp;    if (Sp  < 0.0) Sp  = 0.0;		/* Sp = (d1+dd1) - pitch_engage */
		Sy  = (d2_act + pRec->b) - Ty;    if (Sy  < 0.0) Sy  = 0.0;		/* Sy = (d2+dd2) - yaw_engage */
		Sw  = Sy - Sp * Ly / Lp;          if (Sw  < 0.0) Sw  = 0.0;
		pRec->oa = 1000. * (Sp - Sp0) / Lp;					/* dx1 = 1000*dSp/Lp */
		pRec->ob = 1000. * (Sw - Sw0) / Lw;					/* dx2 = 1000*dSw/Lw */
	}

 	return (0);
}

/* ===========================================
 * tsubAPangAxs - Aperture Angle: Axes to Drives and Motors
 *	oa = x1		(Pitch angle)				<- this is only used to correct x1 when x1 input is < x1_offset
 *	ob = x2		(Yaw angle)				<- this is only used to correct x2 when x2 input is < x2_offset
 *	oa0 = m1	(Pitch pusher motor)
 *	oa1 = d1	(Pitch pusher drive)
 *	ob0 = m2        (Yaw pusher motor)
 *	ob1 = d2        (Yaw pusher drive)
 *	a = x1		(Pitch angle)
 *	b = x2          (Yaw angle)
 *      i = Lp		(Pitch frame level arm: distance between Pitch axis and Pitch prism)
 *      j = Ly		(Yaw frame level vertical arm: distance between Pitch axis and Yaw prism)
 *      k = Lw		(Yaw frame level horizontal arm: distance between Yaw axis and Yaw prism)
 *      l = Tp		(Pitch drive position when it reaches its prism and becomes ready to push)
 *      m = Ty		(Yaw drive position when it reaches its prism and becomes ready to push -- when the Yaw-prism is not displaced by Pitch pusher)
 *	a0 = d1:Offset
 *	a1 = d1:Scale
 *      a2 = d1:ActPos  (Pitch pusher drive actual position)
 *	a3 = x1:Offset	(initial Pitch angle when nothing is pushed)
 *	b0 = d2:Offset
 *	b1 = d2:Scale
 *      b2 = d2:ActPos	(Yaw pusher drive actual position)
 *	b3 = x2:Offset	(initial Yaw angle when nothing is pushed)
 *	nla = 		(1=wo/Offsets)[rel,vel] (0=w/Offsets)[abs,pos]
 */
static long tsubAPangAxs (struct tsubRecord *pRec) {
	double Lp, Ly, Lw, Tp, Ty, Sp, Sy, Sw, Sp0, Sy0, Sw0,
               d1_offset, d1_scale, d1_act, x1_offset, x1, x1_act, dx1,
               d2_offset, d2_scale, d2_act, x2_offset, x2, x2_act, dx2;
	Lp        = pRec->i;
	Ly        = pRec->j;
	Lw        = pRec->k;
	Tp        = pRec->l;
	Ty        = pRec->m;
	d1_offset = pRec->a0;
	d1_scale  = pRec->a1;
	d1_act    = pRec->a2;
	x1_offset = pRec->a3;
	d2_offset = pRec->b0;
	d2_scale  = pRec->b1;
	d2_act    = pRec->b2;
	x2_offset = pRec->b3;

	if (d1_scale == 0.0 || d2_scale == 0.0) {
		printf ("tsubAPangAxs: exit on zero mtr to drv scale\n");
		return (-1);
	}
	if (Lp == 0.0 || Ly  == 0.0 || Lw == 0.0) {
		printf ("tsubAPangAxs: exit on zero Lp, Ly, or Lw\n");
		return (-1);
	}

	if (pRec->nla == 0.0)					   			/* absolute motion */
	{
		x1 = pRec->a;
		x2 = pRec->b;
		if (x1 < x1_offset) {x1=x1_offset;  pRec->oa = x1;}			/* x1 is over the range -- correct! */
		if (x2 < x2_offset) {x2=x2_offset;  pRec->ob = x2;}			/* x2 is over the range -- correct! */
		Sp = (x1 - x1_offset) * Lp / 1000.;					/* Sp = (x1-x1_offset)*Lp/1000 */
		Sw = (x2 - x2_offset) * Lw / 1000.;					/* Sw = (x2-x2_offset)*Lp/1000 */
		Sy = Sw + Sp * Ly / Lp;							/* Sy = Sw+Sp*Ly/Lp */
		pRec->oa1 = Sp + Tp;							/* d1 = Sp + Tp */
		pRec->ob1 = Sy + Ty;							/* d2 = Sy + Ty */
		pRec->oa0 = (pRec->oa1 - d1_offset) / d1_scale;				/* m1=(d1 - offset1)/scale1 */
		pRec->ob0 = (pRec->ob1 - d2_offset) / d2_scale;				/* m2=(d2 - offset2)/scale2 */
	}
	else                                                       			/* relative motion */
	{
		Sp0 = d1_act - Tp;		  if (Sp0 < 0.0) Sp0 = 0.0;		/* Sp0 = d1 - pitch_engage */
		Sy0 = d2_act - Ty;		  if (Sy0 < 0.0) Sy0 = 0.0;		/* Sy0 = d2 - yaw_engage */
		Sw0 = Sy0 - Sp0 * Ly / Lp;	  if (Sw0 < 0.0) Sw0 = 0.0;
                x1_act = x1_offset + 1000. * Sp0 / Lp;					/* x1 = x1_offset + 1000*Sp/Lp */
		x2_act = x2_offset + 1000. * Sw0 / Lw;					/* x2 = x2_offset + 1000*Sw/Lp */
		dx1    = pRec->a;
		dx2    = pRec->b;
		x1     = x1_act + dx1;							/* x1 = x1_act + dx1 */
		x2     = x2_act + dx2;							/* x2 = x2_act + dx2 */
		if (x1 < x1_offset) {x1=x1_offset;  dx1=x1-x1_act;  pRec->oa = dx1;}	/* x1 is over the range -- correct dx1 */
		if (x2 < x2_offset) {x2=x2_offset;  dx2=x2-x2_act;  pRec->ob = dx2;}	/* x2 is over the range -- correct dx2 */
		Sp     = (x1 - x1_offset) * Lp / 1000.;					/* Sp = (x1-x1_offset)*Lp/1000 */
		Sw     = (x2 - x2_offset) * Lw / 1000.;					/* Sw = (x2-x2_offset)*Lp/1000 */
		Sy     = Sw + Sp * Ly / Lp;						/* Sy = Sw+Sp*Ly/Lp */
		pRec->oa1 = Sp - Sp0;							/* d(d1) = d(Sp) */
		pRec->ob1 = Sy - Sy0;							/* d(d2) = d(Sy) */
		pRec->oa0 = pRec->oa1 / d1_scale;                			/* d(m1)=d(d1)/scale1 */
		pRec->ob0 = pRec->ob1 / d2_scale;                			/* d(m2)=d(d2)/scale2 */
	}
	return (0);
}

/* ===========================================
 * tsubAPangSpeed - Speed propagation spreadsheet
 *	oa0 = m1	(Pitch pusher motor)
 *	oa1 = d1	(Pitch pusher drive)
 *	oa2 = x1        (Pitch angle)
 *	ob0 = m2	(Yaw pusher motor)
 *	ob1 = d2	(Yaw pusher drive)
 *	ob2 = x2	(Yaw angle)
 *      oj  = sdis (sdis=1 -- disable record processing)
 *	a = m1_max	(m1 max speed)
 *	a0 = m1		(Pitch pusher motor speed)
 *	a1 = d1		(Pitch pusher drive speed)
 *	a2 = x1		(Pitch angle speed)
 *	a3 = d1:Scale   (Pitch m1 to d1 scale)
 *	b = m2_max	(m2 max speed)
 *	b0 = m2		(Yaw pusher motor speed)
 *	b1 = d2		(Yaw pusher drive speed)
 *	b2 = x2		(Yaw angle speed)
 *	b3 = d2:Scale	(Yaw m2 to d2 scale)
 *      i = Lp		(Pitch frame level arm: distance between Pitch axis and Pitch prism)
 *      k = Lw		(Yaw frame level horizontal arm: distance between Yaw axis and Yaw prism)
 *      nla = Index of input(m1=1, m2=2, d1=11, d2=12, x1=21, x2=22)
 */
static long tsubAPangSpeed (struct tsubRecord *pRec) {
	double prcn = 0.0;
	double m1_max, m2_max, m1, m2, d1, d2, x1, x2,
               Lp, Lw, d1_scale, d2_scale;
	long ifail = 0;

/* Disable next record processing in order to avoid infinite loop.
 * This actually points to a record linked to the SDIS field of the tsub.
 * The SDIS has to be re-enabled before next tsub record call */
   	pRec->oj = 1;                                    /* sdis=1 */

	m1_max    = pRec->a;
        m1        = pRec->a0;
        d1        = pRec->a1;
        x1        = pRec->a2;
	d1_scale  = pRec->a3;
	m2_max    = pRec->b;
        m2        = pRec->b0;
        d2        = pRec->b1;
        x2        = pRec->b2;
	d2_scale  = pRec->b3;
	Lp        = pRec->i;
	Lw        = pRec->k;

  	if (tsubAADebug > 1) printf ("tsubAPangSpeed: called with n=%f \n",pRec->nla);

	if ( m1_max   == 0.0 || m2_max   == 0.0 ||
             d1_scale == 0.0 || d2_scale == 0.0 ||
             Lp       == 0.0 || Lw       == 0.0) {
	   printf ("tsubAPangSpeed: exit on zero calc parameters\n");
	   printf ("tsubAPangSpeed: max1=%5.2f sca1=%g\n",m1_max,d1_scale);
	   printf ("tsubAPangSpeed: max2=%5.2f sca2=%g\n",m2_max,d2_scale);
	   printf ("tsubAPangSpeed:  Lp=%6.3f   Lw=%6.3f\n",Lp,Lw);
           return (-1);
	}

	if      (pRec->nla ==  1.0)  /* ------------------------- m1 speed changed */
	{
	   if ( m1 == 0.0 ) {prcn = 1.0;}		/* if 0, then set to max */
	   else {
	      prcn = fabs( m1 / m1_max );		/* prcn=m1/m1_max */
	   }
	}
	else if (pRec->nla ==  2.0)  /* ------------------------- m2 speed changed */
	{
	   if ( m2 == 0.0 ) {prcn = 1.0;}		/* if 0, then set to max */
	   else {
              prcn = fabs( m2 / m2_max );		/* prcn=m2/m2_max */
	   }
	}
	else if (pRec->nla == 11.0) /* ------------------------- d1 speed changed */
	{
	   if ( d1  < 0.0 ) ifail = -1;			/* x,d speeds are always > 0 */
	   if ( d1 <= 0.0 ) {prcn = 1.0;}		/* if 0, then set to max */
	   else {
	      m1 = d1 / (1000.0 * d1_scale);		/* m1=d1/scale1, m1 speed is in counts/ms */
       	      prcn = fabs( m1 / m1_max );		/* m speed must have same sign as m_max */
	   }
	}
	else if (pRec->nla == 12.0) /* ------------------------- d2 speed changed */
	{
	   if ( d2  < 0.0 ) ifail = -1;			/* x,d speeds are always > 0 */
	   if ( d2 <= 0.0 ) {prcn = 1.0;}		/* if 0, then set to max */
	   else {
	      m2 = d2 / (1000.0 * d2_scale);		/* m2=d2/scale2, m2 speed is in counts/ms */
              prcn = fabs( m2 / m2_max );		/* m speed must have same sign as m_max */
	   }
	}
	else if (pRec->nla == 21.0) /* ------------------------- x1 speed changed */
	{
	   if ( x1  < 0.0 ) ifail = -1;			/* x,d speeds are always > 0 */
	   if ( x1 <= 0.0 ) {prcn = 1.0;}		/* if 0, then set to max */
	   else {
	      d1 = x1 * Lp / 1000.0;			/* d1=x1*Lp/1000, x1 is in milliradians */
	      m1 = d1 / (1000.0 * d1_scale);		/* m1=d1/scale1, m1 speed is in counts/ms */
       	      prcn = fabs( m1 / m1_max );		/* m speed must have same sign as m_max */
	   }
	}
	else if (pRec->nla == 22.0) /* ------------------------- x2 speed changed */
	{
	   if ( x2  < 0.0 ) ifail = -1;			/* x,d speeds are always > 0 */
	   if ( x2 <= 0.0 ) {prcn = 1.0;}		/* if 0, then set to max */
	   else {
	      d2 = x2 * Lw / 1000.0;			/* d2=x2*Lw/1000, x2 is in milliradians */
	      m2 = d2 / (1000.0 * d2_scale);		/* m2=d2/scale2, m2 speed is in counts/ms */
       	      prcn = fabs( m2 / m2_max );		/* m speed must have same sign as m_max */
	   }
	}
	else
	{
	   return (-1);
	}

	if ( prcn < 0.0 || prcn > 1.0 ) {
/* If **anything** is wrong set speed to normal */
          ifail = -1;
	  prcn = 1.0;
	}
	m1 = prcn * m1_max;				/* m1=prcn*m1_max */
	m2 = prcn * m2_max;				/* m2=prcn*m2_max */
	d1 = fabs( 1000.0 * m1 * d1_scale );		/* d1=1000*m1*scale1 (1000 is due to ms->s) */
	d2 = fabs( 1000.0 * m2 * d2_scale );		/* d2=1000*m2*scale2 (1000 is due to ms->s) */
	x1 = fabs( 1000.0 * d1 / Lp );			/* x1=1000*d1/Lp (1000 is due to rad->mrad) */
	x2 = fabs( 1000.0 * d2 / Lw );			/* x2=1000*d2/Lw (1000 is due to rad->mrad) */
	pRec->oa0 = m1;
	pRec->ob0 = m2;
	pRec->oa1 = d1;
	pRec->ob1 = d2;
	pRec->oa2 = x1;
	pRec->ob2 = x2;
  	if (tsubAADebug > 1) printf ("tsubAPangSpeed: m1=%5.2f  m2=%5.2f\n",m1,m2);
	return (ifail);
}

/* ===========================================
 *               Names registration
 *  =========================================== */
static registryFunctionRef tsubAPangRef[] = {
    {"tsubAPang",      (REGISTRYFUNCTION)tsubAPang},
    {"tsubAPangSync",  (REGISTRYFUNCTION)tsubAPangSync},
    {"tsubAPangMtr",   (REGISTRYFUNCTION)tsubAPangMtr},
    {"tsubAPangDrv",   (REGISTRYFUNCTION)tsubAPangDrv},
    {"tsubAPangAxs",   (REGISTRYFUNCTION)tsubAPangAxs},
    {"tsubAPangSpeed", (REGISTRYFUNCTION)tsubAPangSpeed}
};

static void tsubAPangFunc(void) {				/* declare this via registrar in DBD */
    registryFunctionRefAdd(tsubAPangRef,NELEMENTS(tsubAPangRef));
}
epicsExportRegistrar(tsubAPangFunc);

